// uCalcCPP.cpp : Implementation file for uCalc Fast Math Parser 2.96
// Designed to work with 32-bit MS Visual C++ and Borland C++ Builder.
//
// Copyright (c) 1999-2008 by Daniel Corbier.
// Visit www.ucalc.com for additional resources.

#include <stdlib.h>
#include "uCalcCPP.h"

#if defined(uc_Unicode)
   #define uc_LongToString(a, b, c) _ltow(a, b, c)
   #define uc_StringToLong(a) _wtol(a)
   #define uc_STRING_char(a) CW2A(a)
   #define uc_char wchar_t
#else // Multi-byte
   #define uc_LongToString(a, b, c) _ltoa(a, b, c)
   #define uc_StringToLong(a) atol(a)
   #if !defined(__BORLANDC__)
        #define uc_STRING_char(a) a
   #endif
   #define uc_char char
#endif

#if defined(__BORLANDC__)
   #define uc_StrPtr(StringArg) StringArg.c_str()
   #define uc_STRING_char(StringArg) StringArg.c_str()
   #define uc_T(StringArg) StringArg
   typedef String ucSTRING;
#else // VC++
   #define uc_StrPtr(StringArg) StringArg
   #define uc_T(StringArg) _T(StringArg)
   typedef CString ucSTRING;
#endif

#define FuncPtr(Func)     reinterpret_cast<DWORD>(Func)
#define VarPtr(NumberVar) reinterpret_cast<DWORD>(&NumberVar)

long  ucIsInitialized = 0;
long  ucLoadLibFailed = 0;
DWORD ucExtended, ucString, ucLPCSTR, ucFixedString, ucDouble, ucSingle;
DWORD ucLong, ucDword, ucComplex, ucByte, ucTable, ucVoid, ucStack;
DWORD ucSortedList;

UCALC_ uCalc_;
UCALC__ uCalc__;
UCEVALUATE ucEvaluate;
UCEXPRTHREAD ucExprThread;
UCPARAM ucParam;
UCPARAMCOUNT ucParamCount;
UCPARAMHANDLE ucParamHandle;
UCPARAMSTR ucParamStr;
UCRETURN ucReturn;
UCRETURNSTR ucReturnStr_;
UCSETVARIABLEVALUE ucSetVariableValue;
UCSETVARIABLEVALUESTR ucSetVariableValueStr;
UCSTRRETURN ucStrReturn;

DWORD ucInitialize();

DWORD uCalc(long ucCommand, const ucSTRING TextArg = "", DWORD MiscArg1 = 0, DWORD MiscArg2 = 0, DWORD ReturnPtr = 0, DWORD t = 0)
{
   if (ucIsInitialized == ucFalse) ucInitialize();
   return uCalc_(ucCommand, uc_STRING_char(TextArg), MiscArg1, MiscArg2, ReturnPtr, t);
}

DWORD uCalcB(long ucCommand, const DWORD TextArg = 0, DWORD MiscArg1 = 0, DWORD MiscArg2 = 0, DWORD ReturnPtr = 0, DWORD t = 0)
{
   if (ucIsInitialized == ucFalse) ucInitialize();
   return uCalc__(ucCommand, TextArg, MiscArg1, MiscArg2, ReturnPtr, t);
}

ucSTRING uCalcStr(long ucCommand, const ucSTRING TextArg = "", DWORD MiscArg1 = 0, DWORD MiscArg2 = 0, DWORD ReturnPtr = 0, DWORD t = 0)
{
   if (ucIsInitialized == ucFalse) ucInitialize();
   return ucStrReturn(uCalc_(ucCommand, uc_STRING_char(TextArg), MiscArg1, MiscArg2, ReturnPtr, t));
}

DWORD ucAddr(long AddressIndex)
{
   return uCalcB(uc_GetCodeAddr, 0, AddressIndex);
}

DWORD ucDefine(ucSTRING Definition, DWORD Arg1 = 0, DWORD Arg2 = 0, DWORD Arg3 = 0, DWORD t = 0)
{
   return uCalc(uc_Define, uc_StrPtr(Definition), Arg1, Arg2, Arg3, t);
}

DWORD ucDefine(ucSTRING Definition, void *ItemAddress, DWORD Arg2 = 0, DWORD Arg3 = 0, DWORD t = 0)
{
   return uCalc(uc_Define, uc_StrPtr(Definition), reinterpret_cast<DWORD>(ItemAddress), Arg2, Arg3, t);
}

DWORD ucDefineFunction(ucSTRING Definition, void *ItemAddress = 0, DWORD t = 0)
{
   ucSTRING ItemDef;
   uc_char AddressString[15];

   ItemDef = "Func: " + Definition;
   if (ItemAddress != 0)
   {
      uc_LongToString(reinterpret_cast<DWORD>(ItemAddress), AddressString, 10);
      ItemDef = ItemDef + " At " + AddressString;
   }
   return uCalc(uc_Define, uc_StrPtr(ItemDef), 0, 0, 0, t);
}

DWORD ucDefineOperator(ucSTRING Definition, void *ItemAddress = 0, DWORD t = 0)
{
   ucSTRING ItemDef;
   uc_char AddressString[15];

   ItemDef = "Op: " + Definition;
   if (ItemAddress != 0)
   {
      uc_LongToString(reinterpret_cast<DWORD>(ItemAddress), AddressString, 10);
      ItemDef = ItemDef + " At " + AddressString;
   }
   return uCalc(uc_Define, uc_StrPtr(ItemDef), 0, 0, 0, t);
}

DWORD ucDefineSyntax(ucSTRING Definition, long Rank = 1, DWORD t = 0)
{
   return uCalc(uc_DefineSyntax, uc_StrPtr(Definition), Rank, 0, 0, t);
}

DWORD ucDefineVariable(ucSTRING Definition, void *ItemAddress = 0, DWORD t = 0)
{
   ucSTRING ItemDef;
   uc_char AddressString[15];

   ItemDef = "Var: " + Definition;
   if (ItemAddress != 0)
   {
      uc_LongToString(reinterpret_cast<DWORD>(ItemAddress), AddressString, 10);
      ItemDef = ItemDef + " At " + AddressString;
   }
   return uCalc(uc_Define, uc_StrPtr(ItemDef), 0, 0, 0, t);
}

long double ucEval(ucSTRING Expr, DWORD t = 0)
{
   long double Answer;
   Answer = 0;
   uCalc(uc_Eval, uc_StrPtr(Expr), 0, 0, VarPtr(Answer), t);
   return Answer;
}

ucSTRING ucEvalStr(ucSTRING Expr, DWORD ExprType = 0, DWORD t = 0)
{
   return uCalcStr(uc_EvalStr, uc_StrPtr(Expr), ExprType, 0, 0, t);
}

ucSTRING ucEvaluateStr(DWORD Item)
{
   return uCalcStr(uc_EvaluateStr, "", Item);
}

ucSTRING ucExpand(ucSTRING Expr, DWORD t = 0)
{
   return uCalcStr(uc_Expand, uc_StrPtr(Expr), 0, 0, 0, t);
}

ucSTRING ucErrSymbol(DWORD t)
{
   return uCalcStr(uc_GetSymbol, "", 0, 0, 0, t);
}
long ucErrLocation(DWORD t)
{
   return uCalcB(uc_GetLocation, 0, 0, 0, 0, t);
}

ucSTRING ucErrorMessage(DWORD ErrNum = 0, DWORD t = 0)
{
   return uCalcStr(uc_GetErrorMessage, "", ErrNum, 0, 0, t);
}
long ucGetError(DWORD t = 0)
{
   return uCalcB(uc_GetErrorNumber, 0, 0, 0, 0, t);
}
long ucError(DWORD t = 0)
{
   return uCalcB(uc_GetErrorNumber, 0, 0, 0, 0, t);
}
long ucRaiseError(DWORD Expr, long ErrorNum)
{
   return uCalcB(uc_SetErrorNumber, 0, ErrorNum, Expr, 0, ucExprThread(Expr));
}
void ucRaiseErrorMessage(DWORD Expr, ucSTRING ErrMessage)
{
   uCalc(uc_SetErrorMessage, uc_StrPtr(ErrMessage), uc_Err_Dynamically_Defined, 0, 0, ucExprThread(Expr));
   uCalcB(uc_SetErrorNumber, 0, uc_Err_Dynamically_Defined, Expr, 0, ucExprThread(Expr));
}
long ucSetError(long ErrorNum, DWORD t = 0)
{
   return uCalcB(uc_SetErrorNumber, 0, ErrorNum, 0, 0, t);
}

ucSTRING ucGetCodeBlock(ucSTRING Expr, long &i, long ucOption = 0, long t = 0)
{
   return uCalcStr(uc_GetCodeBlock, uc_StrPtr(Expr), 0, ucOption, i, t);
}

ucSTRING ucGetNextItem(ucSTRING Expr, long &i, long ucOption = 0, long t = 0)
{
   return uCalcStr(uc_GetNextItem, uc_StrPtr(Expr), 0, ucOption, i, t);
}

DWORD ucGetItemHandle(ucSTRING SymbolName, DWORD ucProperties = 0, DWORD t = 0)
{
   return uCalc(uc_GetItemData, uc_StrPtr(SymbolName), 0, uc_Handle, ucProperties, t);
}

long ucGetTrigMode()
{
   return uCalc(uc_GetTrigMode);
}

ucSTRING ucInterpret(ucSTRING Expr, DWORD ExprType = 0, DWORD t = 0)
{
   return uCalcStr(uc_EvalStr, uc_StrPtr(Expr), ExprType, 1, 0, t);
}

long ucLicense(long LicenseOption, long L1 = 0, long L2 = 0, long L3 = 0)
{
   static HINSTANCE hDLL = 0;

   if (hDLL == 0)
   {
      hDLL  = LoadLibrary(uc_T("ucFMP296.dll"));

      if (hDLL == NULL)
      {
         ucLoadLibFailed = 1;
         return 0;
         // Host program should check ucLoadLibFailed before continuing
      }

#if defined(__BORLANDC__)
      ucEvaluate    = (UCEVALUATE)GetProcAddress(hDLL, "ucEvaluateExt");
      ucParam       = (UCPARAM)GetProcAddress(hDLL, "ucParamExt");
      ucReturn      = (UCRETURN)GetProcAddress(hDLL, "ucReturnExt");
      ucSetVariableValue = (UCSETVARIABLEVALUE)GetProcAddress(hDLL, "ucSetVariableValueExt");
#else // VC++
      ucEvaluate    = (UCEVALUATE)GetProcAddress(hDLL, "ucEvaluateDbl");
      ucParam       = (UCPARAM)GetProcAddress(hDLL, "ucParamDbl");
      ucReturn      = (UCRETURN)GetProcAddress(hDLL, "ucReturnDbl");
      ucSetVariableValue = (UCSETVARIABLEVALUE)GetProcAddress(hDLL, "ucSetVariableValueDbl");
#endif

      uCalc_        = (UCALC_)GetProcAddress(hDLL, "uCalc");
      uCalc__       = (UCALC__)GetProcAddress(hDLL, "uCalc");
      ucExprThread  = (UCEXPRTHREAD)GetProcAddress(hDLL, "ucExprThread");
      ucParamStr    = (UCPARAMSTR)GetProcAddress(hDLL, "ucParamStrB");
      ucParamHandle = (UCPARAMHANDLE)GetProcAddress(hDLL, "ucParamHandle");
      ucParamCount  = (UCPARAMCOUNT)GetProcAddress(hDLL, "ucParamCount");
      ucReturnStr_  = (UCRETURNSTR)GetProcAddress(hDLL, "ucReturnStr");
      ucSetVariableValueStr = (UCSETVARIABLEVALUESTR)GetProcAddress(hDLL, "ucSetVariableValueStr");
      ucStrReturn   = (UCSTRRETURN)GetProcAddress(hDLL, "ucStrReturnB");
   }

   return uCalc__(uc_License, L1, LicenseOption, L2, L3, 0);
}

DWORD ucNewThread(DWORD ParentThread = 0)
{
   return uCalcB(uc_NewThread, 0, ParentThread);
}

DWORD ucParse(ucSTRING Expr, DWORD ExprType = 0, DWORD t = 0)
{
   return uCalc(uc_Parse, uc_StrPtr(Expr), ExprType, 0, 0, t);
}

DWORD ucReleaseItem(DWORD ucItem)
{
   return uCalcB(uc_ReleaseItem, 0, ucItem);
}

void ucReturnStr(DWORD Expr, const ucSTRING TextArg)
{
	ucReturnStr_(Expr, uc_STRING_char(TextArg));
}

DWORD ucSetErrMsg(long ErrIndex, ucSTRING ErrMessage, DWORD t = 0)
{
   return uCalc(uc_SetErrorMessage, uc_StrPtr(ErrMessage), ErrIndex, 0, 0, t);
}

DWORD ucSetErrorNumber(long ErrorNumber, DWORD t = 0)
{
   return uCalcB(uc_SetErrorNumber, 0, ErrorNumber, 0, 0, t);
}

void ucSetTrigMode(long TrigMode)
{
   uCalcB(uc_SetTrigMode, 0, TrigMode);
}

DWORD ucInitialize()
{
   DWORD ReturnValue;
   uc_char ucProperties[25];
   
   ReturnValue = ucLicense(uc_License_Demo);
   ucIsInitialized = ucTrue;

   ucSetErrMsg(uc_Err_None,                    "No error");
   ucSetErrMsg(uc_Err_Syntax_Error,            "Syntax error");
   ucSetErrMsg(uc_Err_Undefined_Identifier,    "Undefined identifier");
   ucSetErrMsg(uc_Err_Unknown_Error,           "Unknown error");
   ucSetErrMsg(uc_Err_Unrecognized_Pattern,    "Unrecognized pattern");
   ucSetErrMsg(uc_Err_Unrecognized_Command,    "Unrecognized command");
   ucSetErrMsg(uc_Err_Datatype_Mismatch,       "Data type mismatch");
   ucSetErrMsg(uc_Err_CodeBlock_Error,         "Code block delimiter error");
   ucSetErrMsg(uc_Err_Invalid_Argument_Count,  "Invalid number of arguments");
   ucSetErrMsg(uc_Err_Invalid_Definition,      "Invalid definition");
   ucSetErrMsg(uc_Err_Undefined_Callback,      "Undefined callback");
   ucSetErrMsg(uc_Err_ErrMsgAlreadyDefined,    "Error message already defined");
   ucSetErrMsg(uc_Err_ItemCannotBeModified,    "This item cannot be modified");
   ucSetErrMsg(uc_Err_Unrecognized_Member,     "Unrecognized member");
   ucSetErrMsg(uc_Err_Array_Bounds_Exceeded,   "Array bounds exceeded");
   ucSetErrMsg(uc_Err_Float_Denormal_Operand,  "Denormal operand");
   ucSetErrMsg(uc_Err_Float_Divide_By_Zero,    "Division by 0");
   ucSetErrMsg(uc_Err_Float_Inexact_Result,    "Inexact result");
   ucSetErrMsg(uc_Err_Float_Invalid_Operation, "Invalid operation");
   ucSetErrMsg(uc_Err_Float_Overflow,          "Floating point overflow");
   ucSetErrMsg(uc_Err_Float_Stack_Check,       "Floating point stack error");
   ucSetErrMsg(uc_Err_Float_Underflow,         "Floating point underflow");
   ucSetErrMsg(uc_Err_Integer_Divide_By_Zero,  "Integer division by 0");
   ucSetErrMsg(uc_Err_Integer_Overflow,        "Integer overflow");
   ucSetErrMsg(uc_Err_Privileged_Instruction,  "Privileged instruction denied");

   // Extended, Double, and Single must be defined in that order, and prior to other types.
   ucExtended = ucDefine("DataType: Extended ~~ Size: 10 ~~ TypeHandler: #", ucAddr(uc_Extended));
   ucDouble   = ucDefine("DataType: Double   ~~ Size: 8  ~~ TypeHandler: # ~~ ConvertsTo: Extended, #", ucAddr(uc_Double), ucAddr(uc_Convert_Double_Extended));
   ucSingle   = ucDefine("DataType: Single   ~~ Size: 4  ~~ TypeHandler: # ~~ ConvertsTo: Extended, #", ucAddr(uc_Single), ucAddr(uc_Convert_Single_Extended));
   ucLong     = ucDefine("DataType: Long     ~~ Size: 4  ~~ TypeHandler: # ~~ ConvertsTo: Extended, #", ucAddr(uc_Int32),  ucAddr(uc_Convert_Int32_Extended));
   ucDword    = ucDefine("DataType: Dword    ~~ Size: 4  ~~ TypeHandler: # ~~ ConvertsTo: Extended, #", ucAddr(uc_Int32u), ucAddr(uc_Convert_Int32u_Extended));
   ucString   = ucDefine("DataType: String   ~~ Size: 4  ~~ TypeHandler: # ~~ ConvertsTo: LPCSTR", ucAddr(uc_String));
   ucFixedString = ucDefine("DataType: FixedString ~~ TypeHandler: # ~~ ConvertsTo: String ~~ ConvertsFrom: String", ucAddr(uc_String_Fixed));
   ucByte     = ucDefine("DataType: Byte     ~~ Size: 1  ~~ TypeHandler: # ~~ ConvertsTo: Extended, #", ucAddr(uc_Int8u), ucAddr(uc_Convert_Int8u_Extended));
   ucLPCSTR   = ucDefine("DataType: LPCSTR               ~~ TypeHandler: # ~~ ConvertsTo: String", ucAddr(uc_String_LPCSTR));
   ucStack    = ucDefine("DataType: Stack    ~~ Size: 4  ~~ TypeHandler: #", ucAddr(uc_Stack));
   ucSortedList = ucDefine("DataType: SortedList ~~ Size: 4  ~~ TypeHandler: #", ucAddr(uc_Stack));
   ucTable    = ucDefine("DataType: Table    ~~ Size: 4      ~~ TypeHandler: #", ucAddr(uc_Stack));
   ucVoid     = ucDefine("DataType: Void     ~~ Size: 10 ~~ TextData: Ok ~~ TypeHandler: #", ucAddr(uc_Void));

#if !defined(__BORLANDC__)
   ucDefine("DataType: Double ~~ ConvertsFrom: Byte ~~ ConvertsTo: Byte");
   ucDefine("DataType: Double ~~ ConvertsFrom: Dword ~~ ConvertsTo: Dword");
   ucDefine("DataType: Double ~~ ConvertsFrom: Single ~~ ConvertsTo: Single");
#endif

   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Long, #", ucAddr(uc_Convert_Extended_Int32));
   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Dword, #", ucAddr(uc_Convert_Extended_Int32));
   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Double, #", ucAddr(uc_Convert_Extended_Double));
   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Single, #", ucAddr(uc_Convert_Extended_Single));
   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Byte");

   ucDword = ucDefine("DataType: Dword ~~ ConvertsTo: Long");
   ucDefine("DataType: Double ~~ ConvertsTo: Long, #", ucAddr(uc_Convert_Double_Int32));
   ucDefine("DataType: Long ~~ ConvertsTo: Double, #", ucAddr(uc_Convert_Int32_Double));

   ucDefine("Pattern:[\\x00\\x01\\x02]");
   ucDefine("Pattern:\\x03 ~~ \\x03 ~~ DataType: #2", ucQuotedText+ucLiteral, ucString);
   ucDefine("Pattern:\\x05");
   ucDefine("Pattern:\\x06 ~~ \\x07", ucCodeBlock);
   ucDefine("Pattern: :");
   ucDefine("Pattern: :=");
   ucDefine("Pattern:\\[ ~~ \\]", ucCodeBlock);
   ucDefine("Pattern:\\{ ~~ \\} ~~ DataType: #2 ~~ Address: #3", ucCodeBlock, ucStack, ucAddr(uc_Func_Populate_Stack));
   ucDefine("Pattern:[\\q] ~~ [\\q] ~~ DataType: #2", ucQuotedText+ucLiteral, ucString);
   ucDefine("Pattern:['] ~~ ['] ~~ DataType: #2", ucQuotedText+ucLiteral, ucString);
   ucDefine("Pattern:\\x0D\\x0A", ucStatementSep);
   ucDefine("Pattern:;", ucStatementSep);
   ucDefine("Pattern:~t", ucCurrentThread);
   ucDefine("Pattern:~Eval", ucEvalInsert);
   ucDefine("Pattern:~Eval_1 ~~ PassOnce", ucEvalInsert);
   ucDefine("Pattern:~Eval_P", ucEvalDuringParse);
   ucDefine("Pattern:~Expand", ucExpandInsert);
   ucDefine("Pattern:~Expand_1 ~~ PassOnce", ucExpandInsert);
   ucDefine("Pattern:\\.", ucMemberAccess);
   ucDefine("Pattern:\\~\\~");
   ucDefine("Pattern:,", ucArgSeparator);
   ucDefine("Pattern:\\( ~~ \\)", ucCodeBlock);
   ucDefine("Pattern:[ \\t]+", ucWhiteSpace);
   ucDefine("Pattern:[\\|+/*^$&#\\-=%@!`\\\\<>?]+", ucReducible);
#if defined(__BORLANDC__)
   ucDefine("Pattern:([0-9]*\\.)?[0-9]+(e[-+]?[0-9]+)? ~~ DataType: #2", ucLiteral, ucExtended);
#else // VC++
   ucDefine("Pattern:([0-9]*\\.)?[0-9]+(e[-+]?[0-9]+)? ~~ DataType: #2", ucLiteral, ucDouble);
#endif
   
   ucDefine("Pattern:[a-z_][0-9a-z_]* ~~ Name: AlphaNum", ucAlphaNumeric);

   ucDefineSyntax("Op: {Name} '{' {Operand} '}' [As {type}] ::=Param: {Operand} ~~ Name: {Name} ~~ DataType: {type} ~~ Properties: "
                  +(ucSTRING)uc_LongToString(ucPrefix + ucOperator, ucProperties, 10));

   ucDefineSyntax("Op: '{' {Operand} '}' {Name} [As {type}] ::=Param: {Operand} ~~ Name: {Name} ~~ DataType: {type} ~~ Properties: "
                 +(ucSTRING)uc_LongToString(ucPostfix + ucOperator, ucProperties, 10));

   ucDefineSyntax("Op: '{' {LeftOp} '}' {Name} '{' {RightOp} '}' [As {type}] ::=Param: {LeftOp} ~~ Param: {RightOp} ~~ Name: {Name} ~~ DataType: {type} ~~ Properties: "
                  +(ucSTRING)uc_LongToString(ucInfix + ucOperator, ucProperties, 10));

   ucDefineSyntax("Op: {Def} { At {addr} | ':=' {Expr:'.+'} } ::= Op: {Def} {addr: ~~ Address: {addr} ~~ StdCall} {Expr: ~~ FuncBody: {Expr}}");
   ucDefineSyntax("Op: {pValue:' *[0-9]+ '} {Def} ::= Precedence: {pValue} ~~ Op: {Def}");
   ucDefineSyntax("Op: Precedence({op}) ::= Op: ~Eval(Precedence({op}))");
   ucDefineSyntax("Op:: {Def:'.+'} ::= NativeCall ~~ Address: # ~~ Op: {Def}");
   ucDefineSyntax("Op: Native: {Def} At {Addr} ::= NativeCall ~~ Address: {Addr} ~~ Op: {Def}");

   ucDefineSyntax("Var: {VarName:'[ ]*[a-z_][0-9a-z_]*'} [As {Type}] [= {Value}] [At {Addr:' .+'}] ::="
                  "Variable ~~ Name: {VarName}"
                  "{Type: ~~ DataType: {Type}}"
                  "{Value: ~~ Value: {Value}}"
                  "{Addr: ~~ Address: {Addr}}");
   ucDefineSyntax("Var: {VarName:'[ ]*[a-z_][0-9a-z_]*'} As AnyType ::= AnyType ~~ Var: {VarName}");
   ucDefineSyntax("Var: {PassBy:'( +ByRef)|( +ByVal)|( +ByHandle)|( +ByExpr)'} ::= {PassBy} ~~ Var:");
   ucDefineSyntax("Var: {Def} ... ::= Var: {Def} ~~ VariableArgList");
   ucDefineSyntax("Var: ... ::= Var: _ ~~ VariableArgList");
   ucDefineSyntax("Var: {name:' *[_a-z]*'} = {quote:' *[\\x22\\x27]'} ::= Var: {name} As String = {quote}");

   ucDefineSyntax("Const: {Def:'.*'} ::= Var: {Def} ~~ Properties: " + (ucSTRING)uc_LongToString(ucLock, ucProperties, 10));
   ucDefineSyntax("=- ::= = -");

   ucDefineSyntax("Func: {FuncName} ([{ParamList}]) [As {type:' +[a-z]+[a-z0-9]*'}] [{ At {addr} | = {Def:'.+'} }] ::= "
                 "Params: ({ParamList}) ~~ Name: {FuncName} ~~ DataType: {type}"
                 "~~ Properties: "
                 +(ucSTRING)uc_LongToString(ucFunction + ucPrefix, ucProperties, 10)
                 +"{addr: ~~ Address: {addr} ~~ StdCall} ~~ FuncBody: {Def}");

   ucDefineSyntax("Func:: {Def:'.+'} ::= NativeCall ~~ Address: # ~~ Func: {Def}");
   ucDefineSyntax("Func: Native: {Def} At {Addr} ::= NativeCall ~~ Address: {Addr} ~~ Func: {Def}");

   ucDefineSyntax("Params: ([{SingleParam}]) ::= IsParam ~~ { {SingleParam: Var: {SingleParam}} }");
   ucDefineSyntax("Params: ({Variable} = {Default}) ::= Params: ({Variable} ~~ TextData: {Default})");
   ucDefineSyntax("Params: ({Param1}, {OtherParams}) ::= Params: ({Param1}) ~~ Params: ({OtherParams})");
   ucDefineSyntax("Param: {ParamDef} ~~ ::= IsParam ~~ { Var: {ParamDef} } ~~");

   ucDefine("Op:: -{x}    ~~ Precedence: 25", ucAddr(uc_Op_Negate_Extended));
   ucDefine("Op:: {x As String}+{y As String} As String ~~ Precedence: 10", ucAddr(uc_Op_StrConcat));
   ucDefine("Op:: {x As Long}+{y As Long} As Long ~~ Precedence: 10", ucAddr(uc_Op_Add_Int32));

   ucExtended = ucDefine("DataType: Extended ~~ ConvertsTo: Complex, #", ucAddr(uc_Convert_Extended_Complex));
   ucComplex  = ucDefine("DataType: Complex ~~ {Var: Real As Extended} ~~ {Var: Imag As Extended = 1} ~~ TypeHandler: #", ucAddr(uc_Complex));
   ucDefine("Pattern:#i ~~ DataType: #2", ucLiteral, ucComplex);
   ucDefine("Op:: {x As Complex}+{y As Complex} As Complex ~~ Precedence: 10", ucAddr(uc_Op_Add_Complex));
   ucDefine("Op:: {x As Complex}*{y As Complex} As Complex ~~ Precedence: 10", ucAddr(uc_Op_Multiply_Complex));

   ucDefine("Op:: {x}>{y}  As Long ~~ Precedence: 5", ucAddr(uc_Op_GreaterThan_Extended));
   ucDefine("Op:: {x}<{y}  As Long ~~ Precedence: 5", ucAddr(uc_Op_LessThan_Extended));
   ucDefine("Op:: {x}>={y} As Long ~~ Precedence: 5", ucAddr(uc_Op_GreaterOrEqual_Extended));
   ucDefine("Op:: {x}<={y} As Long ~~ Precedence: 5", ucAddr(uc_Op_LessOrEqual_Extended));
   ucDefine("Op:: {x}<>{y} As Long ~~ Precedence: 5", ucAddr(uc_Op_NotEqual_Extended));
   ucDefine("Op:: {x}=={y} As Long ~~ Precedence: 5", ucAddr(uc_Op_Equal_Extended));

   ucDefine("Op:: {x As String}>{y As String}  As Long ~~ Precedence: 5", ucAddr(uc_Op_GreaterThan_Str));
   ucDefine("Op:: {x As String}<{y As String}  As Long ~~ Precedence: 5", ucAddr(uc_Op_LessThan_Str));
   ucDefine("Op:: {x As String}>={y As String} As Long ~~ Precedence: 5", ucAddr(uc_Op_GreaterOrEqual_Str));
   ucDefine("Op:: {x As String}<={y As String} As Long ~~ Precedence: 5", ucAddr(uc_Op_LessOrEqual_Str));
   ucDefine("Op:: {x As String}<>{y As String} As Long ~~ Precedence: 5", ucAddr(uc_Op_NotEqual_Str));
   ucDefine("Op:: {x As String}=={y As String} As Long ~~ Precedence: 5", ucAddr(uc_Op_Equal_Str));

   ucDefine("Op:: {x As Long} And     {y As Long} As Long ~~ Precedence: 4", ucAddr(uc_Op_And_Int32));
   ucDefine("Op:: {x}         AndAlso {ByExpr y}          ~~ Precedence: 4", ucAddr(uc_Op_AndAlso));
   ucDefine("Op:: {x As Long} Or      {y As Long} As Long ~~ Precedence: 3", ucAddr(uc_Op_Or_Int32));
   ucDefine("Op:: {x}         OrElse  {ByExpr y}          ~~ Precedence: 3", ucAddr(uc_Op_OrElse));

#if defined(__BORLANDC__)
   ucDefine("Op:: {x}+{y} ~~ Precedence: 10", ucAddr(uc_Op_Add_Extended));
   ucDefine("Op:: {x}-{y} ~~ Precedence: 10", ucAddr(uc_Op_Subtract_Extended));
   ucDefine("Op:: {x}*{y} ~~ Precedence: 20", ucAddr(uc_Op_Multiply_Extended));
   ucDefine("Op:: {x}/{y} ~~ Precedence: 20", ucAddr(uc_Op_Divide_Extended));
   ucDefine("Op:: {x} Mod {y} ~~ Precedence: 20", ucAddr(uc_Op_Mod_Extended));
   ucDefine("Op:: {x}^{y} ~~ Precedence: 30", ucAddr(uc_Op_PowerOf_ExtToExt));

   ucDefine("Op:: {x As Long}!  ~~ Precedence: 40", ucAddr(uc_Op_Factorial));
   ucDefine("Op:: {x As Long}\\{y As Long} As Long ~~ Precedence: 20", ucAddr(uc_Op_Divide_Int32));

   ucDefine("Func:: Sin(x)", ucAddr(uc_Func_Sin));
   ucDefine("Func:: Cos(x)", ucAddr(uc_Func_Cos));
   ucDefine("Func:: Tan(x)", ucAddr(uc_Func_Tan));
   ucDefine("Func:: Atan(x)", ucAddr(uc_Func_Atan));
   ucDefine("Func:: Abs(x)", ucAddr(uc_Func_Abs_Extended));
   ucDefine("Func:: Exp(x)", ucAddr(uc_Func_Exp));
   ucDefine("Func:: Log(x)", ucAddr(uc_Func_Log));
   ucDefine("Func:: Ceil(x) As Long", ucAddr(uc_Func_Ceil));
   ucDefine("Func:: Sqr(x)", ucAddr(uc_Func_SquareRoot));
#else // VC++
   ucDefine("Op:: {x As Double}+{y As Double} As Double ~~ Precedence: 10", ucAddr(uc_Op_Add_Double));
   ucDefine("Op:: {x As Double}-{y As Double} As Double ~~ Precedence: 10", ucAddr(uc_Op_Subtract_Double));
   ucDefine("Op:: {x As Double}*{y As Double} As Double ~~ Precedence: 20", ucAddr(uc_Op_Multiply_Double));
   ucDefine("Op:: {x As Double}/{y As Double} As Double ~~ Precedence: 20", ucAddr(uc_Op_Divide_Double));
   ucDefine("Op:: {x As Double} Mod {y As Double} As Double ~~ Precedence: 20", ucAddr(uc_Op_Mod_Double));
   ucDefine("Op:: {x As Double}^{y As Double} As Double ~~ Precedence: 30", ucAddr(uc_Op_PowerOf_DblToDbl));

   ucDefine("Op:: {x As Long}!  ~~ Precedence: 40", ucAddr(uc_Op_Factorial));
   ucDefine("Op:: {x As Long}\\{y As Long} As Long ~~ Precedence: 20", ucAddr(uc_Op_Divide_Int32));

   ucDefine("Func:: Sin(x As Double) As Double", ucAddr(uc_Func_Sin_Dbl));
   ucDefine("Func:: Cos(x As Double) As Double", ucAddr(uc_Func_Cos_Dbl));
   ucDefine("Func:: Tan(x As Double) As Double", ucAddr(uc_Func_Tan_Dbl));
   ucDefine("Func:: Atan(x As Double) As Double", ucAddr(uc_Func_Atan_Dbl));
   ucDefine("Func:: Abs(x As Double) As Double", ucAddr(uc_Func_Abs_Dbl));
   ucDefine("Func:: Exp(x As Double) As Double", ucAddr(uc_Func_Exp_Dbl));
   ucDefine("Func:: Log(x As Double) As Double", ucAddr(uc_Func_Log_Dbl));
   ucDefine("Func:: Ceil(x As Double) As Long", ucAddr(uc_Func_Ceil_Dbl));
   ucDefine("Func:: Sqr(x As Double) As Double", ucAddr(uc_Func_SquareRoot_Dbl));
#endif

   ucDefine("Func:: Int(x)", ucAddr(uc_Func_Int));
   ucDefine("Func:: Frac(x)", ucAddr(uc_Func_Frac));
   ucDefine("Func:: Sgn(x)", ucAddr(uc_Func_Sgn));
   ucDefine("Func:: Rand(x = 1)", ucAddr(uc_Func_Rand));
   ucDefine("Func:: Rand(x As Long, y As Long)", ucAddr(uc_Func_Rand));

   ucDefine("Syntax: Max({x}, {y}) ::= IIf({x} > {y}, {x}, {y})");
   ucDefine("Syntax: Max({x}, {y}, {etc}) ::= Max({x}, Max({y}, {etc}))");
   ucDefine("Syntax: Min({x}, {y}) ::= IIf({x} < {y}, {x}, {y})");
   ucDefine("Syntax: Min({x}, {y}, {etc}) ::= Min({x}, Min({y}, {etc}))");

   ucDefine("Func:: BaseConvert(Number As String, Base As Long)", ucAddr(uc_Func_BaseConvert));

   ucDefine("Func:: iif(cond As Long, ByExpr TruePart As String, ByExpr FalsePart As String = '') As String", ucAddr(uc_Func_IIf_Str));
   ucDefine("Func:: iif(cond As Long, ByExpr TruePart, ByExpr FalsePart = 0)", ucAddr(uc_Func_IIf));
   ucDefine("Func: Inf() = 1/0");
   ucDefine("Func: NaN() = 0/0");

   ucDefine("Func:: Error(Message As String)", ucAddr(uc_Raise_Error));
   ucDefine("Func:: ucAddr(index)", ucAddr(uc_Func_ucAddr));
   ucDefine("Func:: ValueAtAddr(DataType As Dword, Address As Dword, Index As Long = 0) ~~ ValueAtAddr", ucAddr(uc_Func_ValueAtAddr));
   ucDefine("Func:: BatchEval(ByHandle Var As Double, ByHandle VarArray As Double, ByHandle ResultArray As Double, "
            "Count As Long, ByExpr Expr As Double) As Void", ucAddr(uc_Func_Batch_Eval));

   ucDefine("Var: NullStr As String At #", ucAddr(uc_NullString));
   ucDefine("Func:: uCalc(Command As Long, TextArg As String = NullStr, MiscArg1 As Long = 0,"
            "MiscArg2 As Long = 0, ReturnPtr As Long = 0, t As Long = 0) As Long", ucAddr(uc_Func_uCalc));

   ucDefine("Func:: uCalcStr(Command As Long, TextArg As String = NullStr, MiscArg1 As Long = 0,"
            "MiscArg2 As Long = 0, ReturnPtr As Long = 0, t As Long = 0) As String", ucAddr(uc_Func_uCalcStr));

   ucDefine("Syntax: Precedence({Item}) ::= uCalc("+(ucSTRING)uc_LongToString(uc_GetItemData, ucProperties, 10)+", {Item}, 0, "+(ucSTRING)uc_LongToString(uc_Precedence, ucProperties, 10)+", 0, ~t)");

   ucDefine("Func:: uc_Loop(ByExpr DoCondition, ByExpr Expression As AnyType, ByExpr LoopCondition) As Void", ucAddr(uc_Func_Loop));
   ucDefine("Func:: uc_For(ByHandle Counter As AnyType, Start, Finish, Step, ByExpr Expr As AnyType) As Void", ucAddr(uc_Func_For));

   ucDefine("Func:: Asc(Character As String, Position As Long = 1) As Long", ucAddr(uc_Func_Asc));
   ucDefine("Func:: Chr(Code As Long) As String", ucAddr(uc_Func_Chr));

   ucDefine("Var: _Number As String = '([0-9]*\\.)?[0-9]+(e[-+]?[0-9]+)?'");
   ucDefine("Var: _HexNumber As String = '([0-9a-z]*\\.)?[0-9a-z]+(e[-+]?[0-9]+)?'");
   ucDefine("Var: _Alpha As String = '[a-z_][0-9a-z_]*'");
   ucDefine("Syntax: {Num:_Number}{Alpha:_Alpha} ::= {Num}*{Alpha}");
   ucDefine("Syntax: {'#b'}{Number:_Number} ::= ~Eval(BaseConvert('{Number}', 2))");
   ucDefine("Syntax: {'#o'}{Number:_Number} ::= ~Eval(BaseConvert('{Number}', 8))");
   ucDefine("Syntax: {'#h'}{Number:_HexNumber} ::= ~Eval(BaseConvert('{Number}', 16))");

   ucDefine("Const: uc_Define = " + (ucSTRING)uc_LongToString(uc_Define, ucProperties, 10));
   ucDefine("Syntax: ucDefine({Def}) ::= uCalc(uc_Define, {Def}, 0, 0, 0, ~t)");

   ucDefine("Syntax: Local([{' +'}]{Variable:'[^=, ]+'}[{etc}], {Expr}) ::= ~Eval_P(ucDefine('Var: {Variable} {etc}')) {Expr} ~Eval_P(ucRename('{Variable}', ''))");
   ucDefine("Syntax: Local({Variable1}, {OtherVariables}, {Expr}) ::= Local({Variable1}, Local({OtherVariables}, {Expr}))");

   ucDefine("Const: uc_DefineSyntax = " + (ucSTRING)uc_LongToString(uc_DefineSyntax, ucProperties, 10));
   ucDefine("Func:: SetSyntaxParams(Params As String, Expr As String, t As Dword = 0) As String", ucAddr(uc_Func_SetSyntaxParams));
   ucDefineSyntax("Macro: {name} [{args:'\\([^\\)]*\\)'}] = {replacement} ::= Syntax: ~Eval(SetSyntaxParams('{args}', '{name}{args} ::= {replacement}'))");

   ucDefine("Const: uc_GetItemData = " + (ucSTRING)uc_LongToString(uc_GetItemData, ucProperties, 10));
   ucDefine("Const: uc_SetItemData = " + (ucSTRING)uc_LongToString(uc_SetItemData, ucProperties, 10));
   ucDefine("Const: uc_SymbolName = " + (ucSTRING)uc_LongToString(uc_SymbolName, ucProperties, 10));
   ucDefine("Const: uc_Handle = " + (ucSTRING)uc_LongToString(uc_Handle, ucProperties, 10));
   ucDefine("Syntax: ucHandle({Item}) ::= uCalc(uc_GetItemData, {Item}, 0, uc_Handle, 0, ~t)");
   ucDefine("Syntax: ucRename({OldName}, {NewName})"
            "::= uCalc(uc_SetItemData, {NewName}, ucHandle({OldName}), uc_SymbolName, 0, ~t)");

   // Arrays
   ucDefine("Const: uc_Address = " + (ucSTRING)uc_LongToString(uc_Address, ucProperties, 10));
   ucDefine("Const: uc_DataType = " + (ucSTRING)uc_LongToString(uc_DataType, ucProperties, 10));
   ucDefine("Const: uc_GetItemData = " + (ucSTRING)uc_LongToString(uc_GetItemData, ucProperties, 10));
   ucDefineSyntax("Address({Item}) ::= ~Eval(uCalc(uc_GetItemData, '{Item}', 0, uc_Address, 0, ~t))");
   ucDefineSyntax("DataType({Item})::= ~Eval(uCalc(uc_GetItemData, '{Item}', 0, uc_DataType, 0, ~t))");
   ucDefine("PassOnce ~~ Syntax: NextDef {etc:'.+'} ::= NextDef {etc}");

   ucDefineSyntax("Var: {'[ ]*'}{ArrayName:_Alpha} ({MaxIndex}) [{etc}]"
    "::= Count: {MaxIndex} ~~ Var: {ArrayName} {etc} ~~ NextDef ~~"
    "Syntax: {ArrayName} ({Index}) ::= ValueAtAddr(DataType({ArrayName}), Address({ArrayName}), {Index})");

   // Assign Value to Variable
   ucDefine("Func:: SetVar(ByHandle Var As AnyType, ByHandle Value As String) As Void", ucAddr(uc_SetVariableValueAnyType));
#if defined(__BORLANDC__)
   ucDefine("Func:: SetVar(ByHandle Var As AnyType, ByHandle Value As Extended) As Void", ucAddr(uc_SetVariableValueAnyType));
   ucDefine("Func:: SetVar(ByHandle Var As Double, ByHandle Value As Double) As Void", ucAddr(uc_SetVariableValueDouble));
   ucDefine("Func:: SetVar(ByHandle Var As Extended, ByHandle Value As Extended) As Void", ucAddr(uc_SetVariableValueExtended));
#else
   ucDefine("Func:: SetVar(ByHandle Var As AnyType, ByHandle Value As Double) As Void", ucAddr(uc_SetVariableValueAnyType));
   ucDefine("Func:: SetVar(ByHandle Var As Double, ByHandle Value As Double) As Void", ucAddr(uc_SetVariableValueDouble));
#endif
#if !defined(__BORLANDC__)
   ucDefine("DefaultType: Double");
#endif

   return ReturnValue;
}
